// 
// Copyright (C) 2001, 2000, 1999,  Rildo Pragana, Jim Noeth, 
//               Andrew Cameron, David Essex.
// Copyright (C) 1993, 1991  Rildo Pragana.
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2.1,
// or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; see the file COPYING.LIB.  If
// not, write to the Free Software Foundation, Inc., 59 Temple Place,
// Suite 330, Boston, MA 02111-1307 USA

//
//
//   COBOL Compiler Run Time Library -- Move Module
//
//

#include "htcoblib.h"


//#define DEBUG_RTS 0
//#define DEBUG_MOVE_RTS 1                    


#define COBMV_9_	0
#define COBMV_B_	6
#define COBMV_C_	12
#define COBMV_F_	18
#define COBMV_X_	24
#define COBMV__9	0
#define COBMV__B	1
#define COBMV__C	2
#define COBMV__E	3
#define COBMV__F	4
#define COBMV__X	5
#define COBMV_MAX	255


/*--------------------------------------------------------------------------*\
 |                                                                          |
 |       void tcob_move( struct fld_desc *FieldDescSrc,                     |
 |                      char            *caSrcBuffer,                       |
 |                      struct fld_desc *FieldDescDest,                     |
 |                      char            *caDestBuffer)                      |
 |                                                                          |
 |   Copy the contents of the field described by FieldDescSrc (data is      |
 |   pointed to by caSrcBuffer) to the field described by FieldDescDest     |
 |   (data will be stored in memory pointed to by caDescBuffer).            |
 |                                                                          |
 |   This routine will appropriately convert data (or format it) as it      |
 |   potentially moves from one data type to another. Truncation and        |
 |   padding will also be done in this routine if field sizes differ.       |
 |   Data types that this routine will handle are:                          |
 |                                                                          |
 |   Type  Source Type            Destination Type       Defined As         |
 |    '9'  Numeric Display    Numeric Display    DTYPE_DISPLAY              |
 |    'D'  Accept Display     Accept Display     DTYPE_ACCEP_DISPLAY        |
 |    'C'  Numeric Packed Decimal Numeric Packed Decimal DTYPE_PACKED       |
 |    'A'  Alpha          Alpha          DTYPE_ALPHA                        |
 |    'X'  AlphaNumeric       AlphaNumeric       DTYPE_ALPHANUMERIC         |
 |    'E'  AlphaNumeric       Numeric Edited     DTYPE_EDITED               |
 |    'G'  AlphaNumeric       AlphaNumeric       DTYPE_GROUP                |
 |    'B'  Numeric Bin Int    Numeric Bin Int    DTYPE_BININT               |
 |    'U'  Floating Point     Floating Point     DTYPE_FLOAT                |
 |                                                                          |
 |   The main entry point is tcob_move, which looks at the data types of    |
 |   the source and destinations, then calls a specific routine to handle   |
 |   those specific data types. The routines are listed below, their name   |
 |   is composed of tcob_move_?_?, where the first question mark is the     |
 |   source data type, and the second question mark is the destination      |
 |   data type. Certain data types are grouped together, such as ALPHA,     |
 |   ALPHANUMERIC, GROUP and EDITED (when it is the source). This reduces   |
 |   the number of routines needed to handle the various data types (the    |
 |   list below is without the prefix tcob_move, i.e. _x_x means:           |
 |   tcob_move_x_x).                                                        |
 |                                                                          |
 |   _x_x  _9_x  _b_x  _f_x  _c_x                                           |
 |   _x_9  _9_9  _b_9  _f_9  _c_9                                           |
 |   _x_c  _9_c  _b_b  _f_b  _c_b                                           |
 |   _x_e  _9_b  _b_c  _f_c  _c_c                                           |
 |   _x_f  _9_e  _b_e  _f_e  _c_e                                           |
 |   _x_b  _9_f  _b_f  _f_f  _c_f                                           |
 |                                                                          |
\*-------------------------------------------------------------------------*/

void tcob_move( struct fld_desc *pfldDesc1, char *caData1,
                struct fld_desc *pfldDesc2, char *caData2 ) {
	int	i;
	extern void (*pMovRtn[])(struct fld_desc *, char *,
			struct fld_desc *, char *);

	switch(pfldDesc1->type) {
		case DTYPE_DISPLAY:
			i = COBMV_9_;
			break;
		case DTYPE_ALPHA:
			i = COBMV_X_;
			break;
		case DTYPE_BININT:
			i = COBMV_B_;
			break;
		case DTYPE_PACKED:
			i = COBMV_C_;
			break;
		case DTYPE_ACCEPT_DISPLAY:
			i = COBMV_X_;
			break;
		case DTYPE_EDITED:
			i = COBMV_X_;
			break;
		case DTYPE_GROUP:
			i = COBMV_X_;
			break;
		case DTYPE_FLOAT:
			i = COBMV_F_;
			break;
		case DTYPE_ALPHANUMERIC:
		case DTYPE_ALPHANUMERICL:
			i = COBMV_X_;
			break;
		default:
			runtime_error(RTERR_INVALID_PIC, pfldDesc1, (void *)caData1);
			return;
	}
	switch(pfldDesc2->type) {
		case DTYPE_DISPLAY:
			i += COBMV__9;
			break;
		case DTYPE_ALPHA:
			i += COBMV__X;
			break;
		case DTYPE_BININT:
			i += COBMV__B;
			break;
		case DTYPE_PACKED:
			i += COBMV__C;
			break;
		case DTYPE_ACCEPT_DISPLAY:
			i += COBMV__X;
			break;
		case DTYPE_EDITED:
			i += COBMV__E;
			break;
		case DTYPE_GROUP:
			i += COBMV__X;
			break;
		case DTYPE_FLOAT:
			i += COBMV__F;
			break;
		case DTYPE_ALPHANUMERIC:
		case DTYPE_ALPHANUMERICL:
			i += COBMV__X;
			break;
		default:
			runtime_error(RTERR_INVALID_PIC, pfldDesc2, (void *)caData2);
			return;
	}
#ifdef DEBUG_MOVE_RTS
	fprintf(stderr, "Debug:  tcob_move type = %d\n", i);
#endif
	pMovRtn[i](pfldDesc1, caData1, pfldDesc2, caData2);
}



/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          runtime_error                                 |
 |                                                                        |
\*------------------------------------------------------------------------*/

void runtime_error(int iErrorNum, struct fld_desc *pField, void *pData) {
	unsigned int		i, j;
	extern int		_iRtErrorNbr[];
	extern char		*_szRtErrorDesc[];

	for(i = 0; _iRtErrorNbr[i] != -1; ++i)
		if(iErrorNum == _iRtErrorNbr[i])
			break;

	fprintf(stderr, "\n\nRun Time Error - ");
	if(_iRtErrorNbr[i] != -1)
		fprintf(stderr, "%s\n", _szRtErrorDesc[i]);
	else
		fprintf(stderr, "Unknown Error\n");

	if(pField) {
		fprintf(stderr, "Field Description: ");
		fprintf(stderr, "len = %ld, type = %c, dec/pscale = %d/%d, all = %d, just_r = %d, signs = %d/%d\n",
			pField->len, pField->type, (char)pField->decimals,
			(char)pField->pscale, pField->all, pField->just_r,
			pField->separate_sign, pField->leading_sign);
	}
	if(!pData) {
		fprintf(stderr, "No data pointer provided\n");
		return;
	}
	
	fprintf(stderr, "Data Dump, Address = %08lX", (unsigned long)pData);
	for(j=0; j<tcob_fldLength(pField); j+=24) {
		fprintf(stderr, "\n%04X: ", (unsigned)j);
		for(i = 0; i < 24; ++i) {
			if((i + j) == tcob_fldLength(pField))
				break;
			fprintf(stderr, "%02X ", ((unsigned char *)pData)[i]);
		}
		fprintf(stderr, "\n      ");
		for(i = 0; i < 24; ++i) {
			if((i + j) == pField->len)
			if((i + j) == tcob_fldLength(pField))
				break;
			if((((unsigned char *)pData)[i] >= ' ') 
			&& (((unsigned char *)pData)[i] < 0x7f))
				fprintf(stderr, " %c ", ((unsigned char *)pData)[i]);
			else
				fprintf(stderr, "   ");
		}
	}
	fprintf(stderr, "\n\n");
}

int _iRtErrorNbr[] = {
	RTERR_INVALID_DATA,
	RTERR_INVALID_PIC,
	-1
};

char *_szRtErrorDesc[] = {
	"Invalid Data Content",
	"Invalid Picture Structure",
	(char *)0
};


/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_move_edited                              |
 |                                                                        |
\*------------------------------------------------------------------------*/

void tcob_move_edited( struct fld_desc *pSrcFld, char *pSrcData,
                       struct fld_desc *pDstFld, char *pDstData ) {
	unsigned int			i, k;
	unsigned int 			iSrcPtr;
	int 				iDstPtr;
	int 				iDecimalPos;
	unsigned int 			iSuppress; /* 0: Not yet suppressing
						    * 1: Currently suppressing
						    * 2: Done suppressing */
	unsigned int 			bIsNegative;
	unsigned int 			bInFraction;
	unsigned int 			bIsAlphaEdited;
	unsigned int			bBlanking;
	extern int			bDecimalComma;

	char				*caWorkData;
	char				*pPic;
	unsigned int			picLen;
	char				cChar;
	char				cFloatChar;
	char				cFillChar;
	char				cInsertion;
	char				cDecimalPoint;
	char				cComma;
	extern char			cCurrencySymbol;

	struct fld_desc			FldWrk;

	memset(&FldWrk, 0, sizeof(struct fld_desc)); /* All values to 0 */
	picLen = tcob_picReqLen(1);
	FldWrk.pic = (char *)malloc(picLen);
	tcob_picCreate (FldWrk.pic, picLen, 'S', 1, NULL);
	FldWrk.just_r = pDstFld->just_r;
	pPic = pDstFld->pic;
	bIsNegative = 0;
	bInFraction = 0;
	bIsAlphaEdited = 0;	/* Assume Numeric-Edited */
	bBlanking = 1;
	cInsertion = '\0';
	if(bDecimalComma) {
		cComma = '.';
		cDecimalPoint = ',';
	} else {
		cComma = ',';
		cDecimalPoint = '.';
	}

	for(i=0; (cChar=tcob_picElemVal(pPic,i)); i++) {
		k = tcob_picElemLen(pPic,i);
		if((cChar == 'X')
		|| (cChar == '9')
		|| (cChar == 'A')) {
			if(cChar != '9')
				bIsAlphaEdited = 1;
			bBlanking = 0;
			FldWrk.len += k;
			if(bInFraction)
				FldWrk.decimals += k;
		} else if((cChar == 'Z')
		       || (cChar == '*')) {
			cInsertion = cChar;
			FldWrk.len += k;
			if(bInFraction)
				FldWrk.decimals += k;
		} else if((cChar == '+')
		       || (cChar == '-')
		       || (cChar == cCurrencySymbol)) {
			if(cInsertion) {
				if(cInsertion == cChar) {
					FldWrk.len += k;
					if(bInFraction)
						FldWrk.decimals += k;
				}
			} else if(k > 1) {
				cInsertion = cChar;
				FldWrk.len += (k - 1);
				if(bInFraction)
					FldWrk.decimals += (k - 1);
			} else {
				unsigned int ii;
				unsigned char c;
				for(ii=i+1; (c=tcob_picElemVal(pPic, ii)); ii++) {
					if (c == cChar) {
						cInsertion = cChar;
						break;
					} else if((c != cDecimalPoint)
					       && (c != cComma)
					       && (c != '0')
					       && (c != '/')
					       && (c != 'B')){
						break;
					}
				}
			}
		} else if((cChar == cDecimalPoint)
		       || (cChar == 'V')) {
			bInFraction = 1;
		}
	}
	if(FldWrk.len == 0) {
		free(FldWrk.pic);
		return;
	}
	caWorkData = (char *)malloc(FldWrk.len);
	if(!caWorkData) {
		free(FldWrk.pic);
		return;
	}
	FldWrk.type = (bIsAlphaEdited) ? DTYPE_ALPHANUMERIC : DTYPE_DISPLAY;
	
	tcob_move(pSrcFld, pSrcData, &FldWrk, caWorkData);

	if(!bIsAlphaEdited)
		bIsNegative = tcob_extract_sign(&FldWrk, caWorkData);
	if(bBlanking) { /* Zero suppression with replacement insertion */
		for(i = 0; i < FldWrk.len; ++i)
			if(caWorkData[i] != '0')
				break;
		if(i == FldWrk.len) {	/* it is zero, so blank it out */
			if (cInsertion == '*') {
				unsigned int j, dp = 0;
				memset(pDstData, '*', pDstFld->len);
				/* reset the decimal point */
				for(j=0; (cChar=tcob_picElemVal(pPic, j)); j++) {
					if (cChar==cDecimalPoint) {
						pDstData[dp] = cChar;
						break;
					}
					dp += tcob_picElemLen(pPic, j);
					if (dp >= pDstFld->len)
						break;
				}
			} else {
				memset(pDstData, ' ', pDstFld->len);
			}
			free(FldWrk.pic);
			return;
		}
	}
	free(FldWrk.pic);

	iSuppress = 0;
	bInFraction = 0;
	cFloatChar = '\0';
	cFillChar = ' ';

	iSrcPtr = 0;
	iDstPtr = 0;
	iDecimalPos = 0;

	/* pPic now becomes an expanded version of the PIC string */
	pPic = tcob_picExpand(pDstFld);
	if(!pPic) {
		free(caWorkData);
		return;
	}
	for(i=0; (cChar=pPic[i]); i++, iDstPtr++) {
		if((cChar == 'X')
		|| (cChar == '9')
		|| (cChar == 'A')) {
			if((iSuppress == 1) /* floating insertion so far */
			&& (cFloatChar)) {
				pDstData[iDstPtr - 1] = cFloatChar;
			}
			pDstData[iDstPtr] = caWorkData[iSrcPtr];
			iSuppress = 2;
			iSrcPtr++;
		} else if((cChar == 'Z')
		       || (cChar == '*')) {
			if (iSuppress == 2) {
				pDstData[iDstPtr] = caWorkData[iSrcPtr];
			} else if (caWorkData[iSrcPtr] != '0') {
				pDstData[iDstPtr] = caWorkData[iSrcPtr];
				iSuppress = 2;
			} else {
				if (cChar == '*')
					cFillChar = '*';
				pDstData[iDstPtr] = cFillChar;
				iSuppress = 1;
			}
			iSrcPtr++;
		} else if((cChar == cCurrencySymbol)
		       || (cChar == '+')
		       || (cChar == '-')) {
			char c;
			if     (cChar == '+') c = (bIsNegative) ? '-' : cChar;
			else if(cChar == '-') c = (bIsNegative) ? cChar : ' ';
			else                  c = cChar;

			if((!cInsertion)
			  ||(cInsertion == 'Z')
			  ||(cInsertion == '*')) {
				pDstData[iDstPtr] = c;
			} else { /* floating insertion */
				if (iSuppress == 1) {
					if (caWorkData[iSrcPtr] != '0') {
						pDstData[iDstPtr-1] = c;
						iSuppress = 2;
					} else {
						pDstData[iDstPtr] = ' ';
					 	iSrcPtr++;
					}
				}

				if (cChar != cInsertion) {
					pDstData[iDstPtr] = c;
				} else if (iSuppress == 2) {
					pDstData[iDstPtr] = caWorkData[iSrcPtr];
					iSrcPtr++;
				} else if (iSuppress == 0) {
					pDstData[iDstPtr] = ' ';
					cFloatChar = c;
					iSuppress = 1;
				}					
			}

		} else if((cChar == cComma) /* simple insertion */
		       || (cChar == '0')
		       || (cChar == '/')
		       || (cChar == 'B')) {
			pDstData[iDstPtr] = (iSuppress == 1)
				? cFillChar
				: ((cChar == 'B') ? ' ' : cChar);

		} else if(cChar == cDecimalPoint) { /* special insertion */
			pDstData[iDstPtr] = cChar;
			if ((iSuppress == 1) && (cFloatChar))
				pDstData[iDstPtr-1] = cFloatChar;
			iSuppress = 2;
		} else if(cChar == 'V') {
			iDstPtr--; /* iDstPtr not incremented */
			iDecimalPos = iDstPtr+1;

		} else if((cChar == 'C')
		       && (pPic[i + 1] == 'R')) {
			if(bIsNegative) {
				pDstData[iDstPtr]   = 'C';
				pDstData[iDstPtr+1] = 'R';
			} else {
				pDstData[iDstPtr]   = ' ';
				pDstData[iDstPtr+1] = ' ';
			}
			i++;
			iDstPtr++;
		} else if((cChar == 'D')
		       && (pPic[i + 1] == 'B')) {
			if(bIsNegative) {
				pDstData[iDstPtr]   = 'D';
				pDstData[iDstPtr+1] = 'B';
			} else {
				pDstData[iDstPtr]   = ' ';
				pDstData[iDstPtr+1] = ' ';
			}
			i++;
			iDstPtr++;
		} else {
			pDstData[iDstPtr] = cChar;
			iSuppress = 2;
		}
	}
	if((iDecimalPos)
	&& (iSuppress == 2)) {
		for(i=iDecimalPos; i < pDstFld->len; i++) {
			if((pDstData[i] >= '0')
			&& (pDstData[i] <= '9'))
				break;
			pDstData[i] = '0';
		}
	}
	free(pPic);
	free(caWorkData);
	return;
}



/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          _DUMP_                                        |
 |                                                                        |
\*------------------------------------------------------------------------*/

void _DUMP_(unsigned char *caData, char *szCount, char *caOut) {
	unsigned int    i, k;
	unsigned char   c;

	k = 0;
	for(i = 0; i < 4; ++i) {
		if(szCount[i] == '\0')
			break;
		k = (k * 10) + (szCount[i] - '0');
	}

	for(i = 0; i < k; ++i) {
		c = (caData[i] >> 4) + '0';
		if (c > '9') c += 7;
		caOut[i * 2] = c;

		c = (caData[i] & 0xf) + '0';
		if (c > '9') c += 7;
		caOut[(i * 2) + 1] = c;
	}
}


/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          _FLDDUMP_                                     |
 |                                                                        |
\*------------------------------------------------------------------------*/

void _FLDDUMP_(struct fld_desc *pField, char *pData, char *szDesc) {
	unsigned int    i, j;

	if(szDesc)
		fprintf(stderr, "%s: ", szDesc);
	if(pField) {
		fprintf(stderr, "len=%ld, type=%c, dec/pscale=%d/%d, ",
			pField->len, pField->type, (char)pField->decimals, (char)pField->pscale);
		fprintf(stderr, "all=%d, jr/bz=%d/%d, signs (l/s)= %d/%d\n",
			pField->all, pField->just_r, pField->blank, pField->leading_sign, pField->separate_sign);
	}
        if (pField->type != DTYPE_GROUP && pField->type != DTYPE_ALPHANUMERICL) {
		fprintf(stderr, "picture:");
		for(i=0; tcob_picElemVal(pField->pic,i); i++) {
			fprintf(stderr, "%c(%d)", tcob_picElemVal(pField->pic,i),
			tcob_picElemLen(pField->pic,i));
		}
		fprintf(stderr, "\n");
	}
	if(!pData) {
		fprintf(stderr, "No data pointer provided\n");
		return;
	}
	
	fprintf(stderr, "Data Dump, Address = %08lX", (unsigned long)pData);
	for(j=0; j<tcob_fldLength(pField); j+=24) {
		fprintf(stderr, "\n%04X: ", (unsigned)j);
		for(i = 0; i < 24; ++i) {
			if((i + j) == tcob_fldLength(pField))
				break;
			fprintf(stderr, "%02X ", ((unsigned char *)pData)[i]);
		}
		fprintf(stderr, "\n      ");
		for(i = 0; i < 24; ++i) {
			if((i + j) == tcob_fldLength(pField))
				break;
			if((((unsigned char *)pData)[i] >= ' ') 
			&& (((unsigned char *)pData)[i] < 0x7f))
				fprintf(stderr, " %c ", ((unsigned char *)pData)[i]);
			else
				fprintf(stderr, "   ");
		}
	}
	fprintf(stderr, "\n\n");
}


/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          initIntValues                                 |
 |  Initialize the values in the _iIntValues_ array.                      |
 |                                                                        |
\*------------------------------------------------------------------------*/

void initIntValues () {
	unsigned int		i;
	extern long long	_iIntValues_[18];

	_iIntValues_[17] = (long long)1;
	for(i = 17; i > 0; --i) 
		_iIntValues_[i - 1] = _iIntValues_[i] * (long long)10;
}


/*------------------------------------------------------------------------*\
 |                                                                        |
 |                     Static Storage Definitions                         |
 |                                                                        |
\*------------------------------------------------------------------------*/

int bDecimalComma = 0;
char cCurrencySymbol = '$';
long long _iIntValues_[18] = {
    (long long)-1, (long long)-1, (long long)-1,
    (long long)-1, (long long)-1, (long long)-1,
    (long long)-1, (long long)-1, (long long)-1,
    (long long)-1, (long long)-1, (long long)-1,
    (long long)-1, (long long)-1, (long long)-1,
    (long long)-1, (long long)-1, (long long)-1
};

void (*pMovRtn[])(struct fld_desc *, char *, 
			struct fld_desc *, char *) = {
		tcob_move_9_9, tcob_move_9_b, tcob_move_9_c, 
		tcob_move_9_e, tcob_move_9_f, tcob_move_9_x,
		tcob_move_b_9, tcob_move_b_b, tcob_move_b_c, 
		tcob_move_b_e, tcob_move_b_f, tcob_move_b_x,
		tcob_move_c_9, tcob_move_c_b, tcob_move_c_c, 
		tcob_move_c_e, tcob_move_c_f, tcob_move_c_x,
		tcob_move_f_9, tcob_move_f_b, tcob_move_f_c, 
		tcob_move_f_e, tcob_move_f_f, tcob_move_f_x,
		tcob_move_x_9, tcob_move_x_b, tcob_move_x_c, 
		tcob_move_x_e, tcob_move_x_f, tcob_move_x_x
		};


#ifdef DEBUG_MOVE_RTS
int	_iMoveLevel_ = 0;
#endif

