// 
// Copyright (C) 2001, 2000, 1999,  Rildo Pragana, Jim Noeth, 
//               Andrew Cameron, David Essex.
// Copyright (C) 1993, 1991  Rildo Pragana.
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2.1,
// or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; see the file COPYING.LIB.  If
// not, write to the Free Software Foundation, Inc., 59 Temple Place,
// Suite 330, Boston, MA 02111-1307 USA

#include "htcoblib.h"

//#define DEBUG_MOVE_RTS 1                    

/*--------------------------------------------------------------------------*\
 |                                                                          |
 |         Move routines where the source field is DTYPE_DISPLAY            |
 |                                                                          |
\*--------------------------------------------------------------------------*/

void tcob_move_9_9(struct fld_desc *pfldDesc1, char *caData1, 
			  struct fld_desc *pfldDesc2, char *caData2) {
	unsigned int			i;
	int				j, k;
	int				bIsNegative;
	int				iSrcDecimals, iSrcPscale;
	int				iDstDecimals, iDstPscale;
	char				caWork[MAX_INTEGERS + MAX_DECIMALS];
	unsigned char			*caSrcData;
	unsigned char			*caDstData;
	struct fld_desc			fSrcWork;
	struct fld_desc			fDstWork;

#ifdef DEBUG_MOVE_RTS
	{
		extern int _iMoveLevel_;
		fprintf(stderr, "Debug:  Enter tcob_move_9_9 (%d)\n", _iMoveLevel_++);
	}
#endif

	memcpy(&fSrcWork, pfldDesc1, sizeof(struct fld_desc));
	memcpy(&fDstWork, pfldDesc2, sizeof(struct fld_desc));
	caSrcData = (unsigned char *)caData1;
	caDstData = (unsigned char *)caData2;
	iSrcDecimals = (char)fSrcWork.decimals;		/* do not remove (char) */
	iDstDecimals = (char)fDstWork.decimals;		/* do not remove (char) */
	iSrcPscale = (char)fSrcWork.pscale;		/* do not remove (char) */
	iDstPscale = (char)fDstWork.pscale;		/* do not remove (char) */

	if(tcob_picElemVal(fSrcWork.pic,0) == 'S') {	/* Variable is signed */
		if(fSrcWork.separate_sign) {
			char cSign;
			if(fSrcWork.leading_sign) {
				cSign = caSrcData[0];
				++caSrcData;		/* point past the sign */
			} else
				cSign = caSrcData[fSrcWork.len - 1];
			fSrcWork.len --;		/* subtract length of sign */
			bIsNegative = (cSign == '-') ? 1 : 0;
		} else					/* Sign is non-separate */
			bIsNegative = tcob_extract_sign(&fSrcWork, (char *)caSrcData);
	} else				/* Variable is unsigned */
		bIsNegative = 0;
	
	if((tcob_picElemVal(fDstWork.pic,0) == 'S') && (fDstWork.separate_sign)) {
		if(fDstWork.leading_sign) 
			++caDstData;		/* point past the sign area */
		fDstWork.len --;		/* subtract length for sign */
	}

	if(fSrcWork.all) {			/* this is a 'move all' */
		i = 0;
		j = 0;
		k = fSrcWork.len;
		while(i < fDstWork.len) {
			caDstData[i++] = caSrcData[j++];
			if(j == k)
				j = 0;
		}
		if((tcob_picElemVal(fSrcWork.pic,0) == 'S') && (!fSrcWork.separate_sign))
			tcob_put_sign(&fSrcWork, (char *)caSrcData, bIsNegative);
		if(tcob_picElemVal(fDstWork.pic,0) == 'S') {
			if(fDstWork.separate_sign) {
				int iLoc = (fDstWork.leading_sign) ? 0 : fDstWork.len;
				caData2[iLoc] = (bIsNegative) ? '-' : '+';
			} else
				tcob_put_sign(&fDstWork, (char *)caDstData, bIsNegative);
		}
#ifdef DEBUG_MOVE_RTS
	{
		extern int _iMoveLevel_;
		fprintf(stderr, "Debug:  Leaving tcob_move_9_9 (%d)\n", --_iMoveLevel_);
		_FLDDUMP_(pfldDesc1, caData1, "Source");
		_FLDDUMP_(pfldDesc2, caData2, "Dest");
	}
#endif
		return;
	}						/* END OF MOVE ALL CODE */

	memset(caWork, '0', sizeof(caWork)); /* move src to 9(18)V9(18) */
	if(iSrcPscale < 0)		/* integer scaling */
		j = MAX_DIGITS - (iSrcPscale * -1) - fSrcWork.len;
	else if(iSrcDecimals > (int)fSrcWork.len)	/* fractional scaling */
		j = MAX_DIGITS + iSrcDecimals - fSrcWork.len;
	else
		j = MAX_DIGITS - (fSrcWork.len - fSrcWork.decimals);
	
	memmove(&caWork[j], caSrcData, fSrcWork.len);
	if((tcob_picElemVal(fSrcWork.pic,0) == 'S') && (!fSrcWork.separate_sign)) 
		tcob_put_sign(&fSrcWork, (char *)caSrcData, bIsNegative);

	if(iDstPscale < 0)		/* integer scaling */
		j = MAX_DIGITS - (iDstPscale * -1) - fDstWork.len;
	else if(iDstDecimals > (int)fDstWork.len)	/* fractional scaling */
		j = MAX_DIGITS + iDstDecimals - fDstWork.len;
	else
		j = MAX_DIGITS - (fDstWork.len - fDstWork.decimals);
	
	memmove(caDstData, &caWork[j], fDstWork.len);
	if(tcob_picElemVal(fDstWork.pic,0) == 'S') {
		if(fDstWork.separate_sign) {
			int iLoc = (fDstWork.leading_sign) ? 0 : fDstWork.len;
			caData2[iLoc] = (bIsNegative) ? '-' : '+';
		} else
			tcob_put_sign(&fDstWork, (char *)caDstData, bIsNegative);
	}
#ifdef DEBUG_MOVE_RTS
	{
		extern int _iMoveLevel_;
		fprintf(stderr, "Debug:  Leaving tcob_move_9_9 (%d)\n", --_iMoveLevel_);
		_FLDDUMP_(pfldDesc1, caData1, "Source");
		_FLDDUMP_(pfldDesc2, caData2, "Dest");
	}
#endif
	return;
}


void tcob_move_9_b(struct fld_desc *pfldDesc1, char *caData1, 
			  struct fld_desc *pfldDesc2, char *caData2) {
	int		i, j, k, m;
	int		bIsNegative;
	int		iSrcDecimals, iSrcPscale;
	int		iDstDecimals;
	unsigned int	iDestLen;
	int		iSrcLen;
	int		scalef;
	char		caWork[MAX_INTEGERS + MAX_DECIMALS];
	unsigned int	picLen;
	unsigned char	*caSrcData;
	struct fld_desc	fSrcWork;

	long long	iLongField;
	extern long long _iIntValues_[18];

#ifdef DEBUG_MOVE_RTS
	{
		extern int _iMoveLevel_;
		fprintf(stderr, "Debug:  Enter tcob_move_9_b (%d)\n", _iMoveLevel_++);
	}
#endif
	if(_iIntValues_[0] == (long long)-1)	/* first time called */
		initIntValues();
	memcpy(&fSrcWork, pfldDesc1, sizeof(struct fld_desc));
	caSrcData = (unsigned char *)caData1;
	iSrcDecimals = (char)fSrcWork.decimals;		/* do not remove (char) */	
	iSrcPscale = (char)fSrcWork.pscale;		/* do not remove (char) */

	if(tcob_picElemVal(fSrcWork.pic,0) == 'S') {	/* Variable is signed */
		if(fSrcWork.separate_sign) {
			char cSign;
			if(fSrcWork.leading_sign) {
				cSign = caSrcData[0];
				++caSrcData;		/* point past the sign */
			} else
				cSign = caSrcData[fSrcWork.len - 1];
			fSrcWork.len --;		/* subtract length of sign */
			bIsNegative = (cSign == '-') ? 1 : 0;
		} else					/* Sign is non-separate */
			bIsNegative = tcob_extract_sign(&fSrcWork, (char *)caSrcData);
	} else				/* Variable is unsigned */
		bIsNegative = 0;

	/* iDestLen = Number of '9's in the destination's PIC */
	iDestLen = 0;
	for(i=0; tcob_picElemVal(pfldDesc2->pic,i); i++) {
		if(tcob_picElemVal(pfldDesc2->pic,i) == '9')
			iDestLen += tcob_picElemLen(pfldDesc2->pic,i);
	}
	iSrcLen = fSrcWork.len;
	if(iSrcDecimals > 0)
		iSrcLen -= iSrcDecimals;	/* don't want fraction part */

	if(fSrcWork.all) {			/* this is a 'move all' */
		i = 0;
		j = 0;
		k = iSrcLen;
		while(i < MAX_DIGITS) { 
			caWork[i++] = caSrcData[j++];
			if(j == k)
				j = 0;
		}
		fSrcWork.len = MAX_DIGITS;
		fSrcWork.decimals = 0;
		fSrcWork.pscale = 0;
		picLen = tcob_picReqLen(1);
		fSrcWork.pic = (char *)malloc(picLen);
		tcob_picCreate (fSrcWork.pic, picLen, '9', MAX_DIGITS, NULL);
		caSrcData = (unsigned char *)caWork;
	}

        iSrcDecimals = pfldDesc1->decimals;
        iDstDecimals = pfldDesc2->decimals;
        iSrcLen      = pfldDesc1->len;
	if ((tcob_picElemVal(pfldDesc1->pic,0) == 'S') && (pfldDesc1->separate_sign))
		iSrcLen --;
        iDestLen     = tcob_picCompLength(pfldDesc2);

        j = ((iSrcLen - iSrcDecimals) - (iDestLen - iDstDecimals));
	if (j < 0)
		j = 0;
        
	m = 17;
	k = iSrcLen - 1;
	iLongField = 0;

	for(i=0; i<=k-j; i++) {
		iLongField += (_iIntValues_[m-i] * (long long)(caSrcData[k-i] - '0'));
#ifdef DEBUG_MOVE_RTS
	       fprintf(stderr, "Debug: tcob_move_9_b 4: iLongField=%lld, m=%d, j=%d, k=%d;\n", 
	                       iLongField, m, j, k);
#endif
	}

	// and we scale
	if (iSrcDecimals != iDstDecimals) {
		scalef = iDstDecimals - iSrcDecimals;
 		if(scalef < 0) {
			for(i=0; i < -scalef; i++)
				iLongField /= (long long)10;
		} else {
			for(i=0; i < scalef; i++)
				iLongField *= (long long)10;
		}
#ifdef DEBUG_MOVE_RTS
	       fprintf(stderr, "Debug: tcob_move_9_b 4s: iLongField=%lld, s=%d;\n", 
	                       iLongField, scalef);
#endif
	}
	if(iSrcPscale < 0) {
		for(i=0; i < -iSrcPscale; i++)
			iLongField *= (long long)10;
	}
	if((tcob_picElemVal(pfldDesc2->pic,0) == 'S') && (bIsNegative))	/* If negative */
		iLongField *= (long long)-1;

	switch(pfldDesc2->len) {
		case 1:
			*(char *)caData2 = iLongField;
			break;
		case 2:
			*(short int *)caData2 = iLongField;
			break;
		case 4:
			*(int *)caData2 = iLongField;
			break;
		case 8:
			*(long long *)caData2 = iLongField;
			break;
	}

	/* Put removed sign back on the working copy of the source */
	if((tcob_picElemVal(fSrcWork.pic,0) == 'S') && (!fSrcWork.separate_sign))
		tcob_put_sign(&fSrcWork, (char *)caSrcData, bIsNegative);

#ifdef DEBUG_MOVE_RTS
	{
		extern int _iMoveLevel_;
		fprintf(stderr, "Debug:  Leaving tcob_move_9_b (%d)\n", --_iMoveLevel_);
		_FLDDUMP_(pfldDesc1, caData1, "Source");
		_FLDDUMP_(pfldDesc2, caData2, "Dest");
	}
#endif
	if(fSrcWork.all)
		free(fSrcWork.pic);
	return;
}




void tcob_move_9_c(struct fld_desc *pfldDesc1, char *caData1, 
			  struct fld_desc *pfldDesc2, char *caData2) {
	unsigned int			i;
	int				j, k;
	int				bIsNegative;
	int				iSrcDecimals, iSrcPscale;
	int				iDstDecimals, iDstPscale;
	char				caWork[MAX_INTEGERS + MAX_DECIMALS];
	unsigned char			*caSrcData;
	unsigned char			*caDstData;
	unsigned char			iDigit;
	struct fld_desc			fSrcWork;
	struct fld_desc			fDstWork;

#ifdef DEBUG_MOVE_RTS
	{
		extern int _iMoveLevel_;
		fprintf(stderr, "Debug:  Enter tcob_move_9_c (%d)\n", _iMoveLevel_++);
	}
#endif
	memcpy(&fSrcWork, pfldDesc1, sizeof(struct fld_desc));
	memcpy(&fDstWork, pfldDesc2, sizeof(struct fld_desc));
	caSrcData = (unsigned char *)caData1;
	caDstData = (unsigned char *)caData2;
	iSrcDecimals = (char)fSrcWork.decimals;		/* do not remove (char) */
	iDstDecimals = (char)fDstWork.decimals;		/* do not remove (char) */
	iSrcPscale = (char)fSrcWork.pscale;		/* do not remove (char) */
	iDstPscale = (char)fDstWork.pscale;		/* do not remove (char) */
	
	if(tcob_picElemVal(fSrcWork.pic,0) == 'S') {	/* Variable is signed */
		if(fSrcWork.separate_sign) {
			char cSign;
			if(fSrcWork.leading_sign) {
				cSign = caSrcData[0];
				++caSrcData;		/* point past the sign */
			} else
				cSign = caSrcData[fSrcWork.len - 1];
			fSrcWork.len --;		/* subtract length of sign */
			bIsNegative = (cSign == '-') ? 1 : 0;
		} else					/* Sign is non-separate */
			bIsNegative = tcob_extract_sign(&fSrcWork, (char *)caSrcData);
	} else				/* Variable is unsigned */
		bIsNegative = 0;

	if(fSrcWork.all) {			/* this is a 'move all' */
		i = 0;
		j = 0;
		k = fSrcWork.len;
		while(i < (MAX_DIGITS * 2)) { 
			caWork[i++] = caSrcData[j++];
			if(j == k)
				j = 0;
		}
	} else {
		memset(caWork, '0', sizeof(caWork)); /* move src to 9(18)V9(18) */
		if(iSrcPscale < 0)		/* integer scaling */
			j = MAX_DIGITS - fSrcWork.len + (iSrcPscale * -1);
		else if(iSrcDecimals > (int)fSrcWork.len)	/* fractional scaling */
			j = MAX_DIGITS + iSrcDecimals - fSrcWork.len;
		else
			j = MAX_DIGITS - (fSrcWork.len - fSrcWork.decimals);
		
		memmove(&caWork[j], caSrcData, fSrcWork.len);
		if((tcob_picElemVal(fSrcWork.pic,0) == 'S') && (!fSrcWork.separate_sign)) 
			tcob_put_sign(&fSrcWork, (char *)caSrcData, (char)bIsNegative);
	}

	if(iDstPscale < 0)		/* integer scaling */
		j = MAX_DIGITS - fDstWork.len - (iDstPscale * -1);
	else if(iDstDecimals > (int)fDstWork.len)	/* fractional scaling */
		j = MAX_DIGITS + iDstDecimals - fDstWork.len;
	else
		j = MAX_DIGITS - (fDstWork.len - fDstWork.decimals);

	memset (caDstData, 0, (fDstWork.len/2)+1);
	k = ((fDstWork.len & 1) == 0) ? 1 : 0;
	for(i = 0; i < fDstWork.len; ++i) {
		iDigit = caWork[j + i] - '0';
		caDstData[k / 2] |= ((k & 1) ? (iDigit) : (iDigit << 4));
		++k;
	}
	if(tcob_picElemVal(fDstWork.pic,0) == 'S')
		iDigit = (bIsNegative) ? 0x0D : 0x0C;
	else
		iDigit = 0x0F;
	caDstData[fDstWork.len / 2] |= iDigit;
#ifdef DEBUG_MOVE_RTS
	{
		extern int _iMoveLevel_;
		fprintf(stderr, "Debug:  Leaving tcob_move_9_c (%d)\n", --_iMoveLevel_);
		_FLDDUMP_(pfldDesc1, caData1, "Source");
		_FLDDUMP_(pfldDesc2, caData2, "Dest");
	}
#endif
	return;
}



void tcob_move_9_e(struct fld_desc *pfldDesc1, char *caData1, 
			  struct fld_desc *pfldDesc2, char *caData2) {
#ifdef DEBUG_MOVE_RTS
	{
		extern int _iMoveLevel_;
		fprintf(stderr, "Debug:  Enter tcob_move_9_e (%d)\n", _iMoveLevel_++);
	}
#endif
	tcob_move_edited(pfldDesc1, caData1, pfldDesc2, caData2);
#ifdef DEBUG_MOVE_RTS
	{
		extern int _iMoveLevel_;
		fprintf(stderr, "Debug:  Leaving tcob_move_9_e (%d)\n", --_iMoveLevel_);
		_FLDDUMP_(pfldDesc1, caData1, "Source");
		_FLDDUMP_(pfldDesc2, caData2, "Dest");
	}
#endif
	return;
}





void tcob_move_9_f(struct fld_desc *pfldDesc1, char *caData1, 
			  struct fld_desc *pfldDesc2, char *caData2) {
	int				i, j, k;
	int				bIsNegative;
	int				iSrcDecimals, iSrcPscale;
	char				caWork[MAX_INTEGERS + MAX_DECIMALS + 2];
	unsigned char			*caSrcData;
	struct fld_desc			fSrcWork;
	double				dWork;

#ifdef DEBUG_MOVE_RTS
	{
		extern int _iMoveLevel_;
		fprintf(stderr, "Debug:  Enter tcob_move_9_f (%d)\n", _iMoveLevel_++);
	}
#endif
	memcpy(&fSrcWork, pfldDesc1, sizeof(struct fld_desc));
	caSrcData = (unsigned char *)caData1;
	iSrcDecimals = (char)fSrcWork.decimals;		/* do not remove (char) */
	iSrcPscale = (char)fSrcWork.pscale;		/* do not remove (char) */
	
	if(tcob_picElemVal(fSrcWork.pic,0) == 'S') {	/* Variable is signed */
		if(fSrcWork.separate_sign) {
			char cSign;
			if(fSrcWork.leading_sign) {
				cSign = caSrcData[0];
				++caSrcData;		/* point past the sign */
			} else
				cSign = caSrcData[fSrcWork.len - 1];
			fSrcWork.len --;		/* subtract length of sign */
			bIsNegative = (cSign == '-') ? 1 : 0;
		} else					/* Sign is non-separate */
			bIsNegative = tcob_extract_sign(&fSrcWork, (char *)caSrcData);
	} else				/* Variable is unsigned */
		bIsNegative = 0;

	if(fSrcWork.all) {			/* this is a 'move all' */
		i = 0;
		j = 0;
		k = fSrcWork.len;
		while(i < (MAX_DIGITS * 2)) { 
			caWork[i++] = caSrcData[j++];
			if(j == k)
				j = 0;
		}
	} else {
		memset(caWork, '0', sizeof(caWork)); /* move src to 9(18)V9(18) */
		if(iSrcPscale < 0)		/* integer scaling */
			j = MAX_DIGITS - fSrcWork.len - (iSrcPscale * -1);
		else if(iSrcDecimals > (int)fSrcWork.len)	/* fractional scaling */
			j = MAX_DIGITS + iSrcDecimals - fSrcWork.len;
		else
			j = MAX_DIGITS - (fSrcWork.len - fSrcWork.decimals);

		memmove(&caWork[j], caSrcData, fSrcWork.len);
		if((tcob_picElemVal(fSrcWork.pic,0) == 'S') && (!fSrcWork.separate_sign)) 
			tcob_put_sign(&fSrcWork, (char *)caSrcData, bIsNegative);
	}
	/* re-form to create a null-terminated string with a decimal point */
	memmove(caWork + MAX_INTEGERS + 1, caWork + MAX_INTEGERS, MAX_DECIMALS);
	caWork[MAX_INTEGERS] = '.';
	caWork[MAX_INTEGERS + MAX_DECIMALS + 1] = '\0';
	
	dWork = atof(caWork);
	if(bIsNegative)
		dWork *= (double)-1.0;
	if(pfldDesc2->len == 4)
		*(float *)caData2 = (float)dWork;
	else
		*(double *)caData2 = dWork;
#ifdef DEBUG_MOVE_RTS
	{
		extern int _iMoveLevel_;
		fprintf(stderr, "Debug:  Leaving tcob_move_9_f (%d)\n", --_iMoveLevel_);
		_FLDDUMP_(pfldDesc1, caData1, "Source");
		_FLDDUMP_(pfldDesc2, caData2, "Dest");
	}
#endif
	return;
}
		


void tcob_move_9_x(struct fld_desc *pfldDesc1, char *caData1, 
			  struct fld_desc *pfldDesc2, char *caData2) {
	unsigned int		i;
	int			j, k, offset;
	int			bIsNegative;
	int			iSrcDecimals;
	unsigned char		*caSrcData;
	struct fld_desc		fSrcWork;

#ifdef DEBUG_MOVE_RTS
	{
		extern int _iMoveLevel_;
		fprintf(stderr, "Debug:  Enter tcob_move_9_x (%d)\n", _iMoveLevel_++);
	}
#endif

	memcpy(&fSrcWork, pfldDesc1, sizeof(struct fld_desc));
	caSrcData = (unsigned char *)caData1;
	iSrcDecimals = (char)fSrcWork.decimals;		/* do not remove (char) */
	offset = 0;

	if(tcob_picElemVal(fSrcWork.pic,0) == 'S') {
		if(fSrcWork.separate_sign) {
			--fSrcWork.len;
			if (fSrcWork.leading_sign)
			       offset = 1;
		} else
			bIsNegative = tcob_extract_sign(&fSrcWork, (char *)caSrcData);
	}
        
	if(pfldDesc1->all) {	/* this is a 'move all' */
		i = 0;
		j = offset;
		k = fSrcWork.len;

		while(i < pfldDesc2->len) {
			caData2[i++] = caData1[j++];
			if(j == k)
				j = offset;
		}

#ifdef DEBUG_MOVE_RTS
	{
		extern int _iMoveLevel_;
		fprintf(stderr, "Debug:  Leaving tcob_move_9_x (%d)\n", --_iMoveLevel_);
		_FLDDUMP_(pfldDesc1, caData1, "Source");
		_FLDDUMP_(pfldDesc2, caData2, "Dest");
	}
#endif
		return;
	}	
	/* END OF MOVE ALL CODE */

	memset(caData2, ' ', pfldDesc2->len); 
	if(fSrcWork.len > pfldDesc2->len) {
		j = 0;
		i = pfldDesc2->len;
	} else if(pfldDesc2->just_r) {
		j = pfldDesc2->len - fSrcWork.len;
		i = fSrcWork.len;
	} else {
		j = 0;
		i = fSrcWork.len;
	}
        memmove(&caData2[j], (caData1 + offset), i);

#ifdef DEBUG_MOVE_RTS
	{
		extern int _iMoveLevel_;
		fprintf(stderr, "Debug:  Leaving tcob_move_9_x (%d)\n", --_iMoveLevel_);
		_FLDDUMP_(pfldDesc1, caData1, "Source");
		_FLDDUMP_(pfldDesc2, caData2, "Dest");
	}
#endif
	return;
}


